# coding:utf-8

__all__ = ['BaseKey']

__metaclass__ = type

from ooredis.client import get_client
from ooredis.type_case import GenericTypeCase

class BaseKey:

    """ 
    所有其他 key 对象的基类。
    """

    def __init__(self, name, client=None, type_case=GenericTypeCase):
        """ 
        指定 key 名和客户端，以及 type_case 。

        Args:
            name: Redis key 的名字
            client: 客户端，默认为全局客户端。
            type_case: 类型转换类

        Time:
            O(1)

        Returns:
            None

        Raises:
            None
        """
        self.name = name
        self._client = client or get_client()
        self.encode = type_case.encode
        self.decode = type_case.decode


    def __eq__(self, other):
        """ 
        判断两个 Key 对象是否相等。

        Args:
            other: 另一个 Key 对象。

        Time:
            O(1)

        Returns:
            bool: 相等返回 True ，否则返回 False 。

        Raises:
            None
        """
        return self.name == other.name


    @property
    def _represent(self):
        """ 
        返回 Key 对象在 Redis 中的表示(底层的实现类型)。

        Args:
            None

        Time:
            O(1)

        Returns: 
            redis的类型定义在REDIS_TYPE常量中，包含以下：
            'none' : 值不存在
            'string' : 字符串
            'list' : 列表
            'set' : 集合
            'zset' : 有序集
            'hash' : 哈希表

        Raises:
            None
        """
        return self._client.type(self.name)


    @property
    def ttl(self):
        """ 
        返回 key 的生存时间。

        Args:
            None

        Time:
            O(1)

        Returns: 
            None: key 不存在，或 key 没有设置生存时间。
            ttl_in_second: 以秒为单位的生存时间值。

        Raises:
            None
        """
        return self._client.ttl(self.name)


    @property
    def exists(self):
        """ 
        检查 key 是否存在。

        Args:
            None

        Time:
            O(1)

        Returns:
            bool: key 存在返回 True ，否则为 False 。

        Raises:
            None
        """
        return self._client.exists(self.name)


    def delete(self):
        """ 
        删除 key 。

        Args:
            None

        Time:
            O(1)

        Returns:
            None

        Raises:
            None
        """
        self._client.delete(self.name)


    def expire(self, second):
        """ 
        为 key 设置生存时间。

        Args:
            second: 以秒为单位的生存时间。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对不存在的 key 进行设置时抛出。
        """
        # redis-py 对不存在的 key 进行 expire 时返回false，
        # 这里抛出一个异常。
        if not self.exists:
            raise TypeError

        self._client.expire(self.name, second)


    def expireat(self, unix_timestamp):
        """ 
        为 key 设置生存时间，以一个 unix 时间戳为终止时间。

        Args:
            unix_timestamp: 以 unix 时间戳为格式的生存时间。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 对一个不存在的 key 进行操作时抛出。
        """
        # redis-py 对不存在的 key 进行 expireat 时返回False，
        # 这里抛出一个异常。
        if not self.exists:
            raise TypeError

        self._client.expireat(self.name, unix_timestamp)


    def persist(self):
        """ 
        移除 key 的生存时间

        Args:
            None

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 对一个不存在的 key 进行操作时抛出
        """
        # redis-py 对不存在的 key 进行 persist 返回-1
        # 这里抛出一个异常
        if not self.exists:
            raise TypeError

        self._client.persist(self.name)



# coding: utf-8

import json

class JsonTypeCase:

    """ 
    用于将 python 对象转化为 JSON 对象。 
    """

    @staticmethod
    def encode(value):
        """ 
        将值转成 JSON 对象，
        如果值不能转成 JSON 对象，抛出 TypeError 异常。
        """
        return json.dumps(value)
    
    @staticmethod
    def decode(value):
        """ 
        将 Json 对象转回原来的类型。
        如果转换失败，抛出 TypeError 。
        """
        if value is None:
            return None

        return json.loads(value)



# coding: utf-8

class StringTypeCase:

    """ 
    处理字符串类型(str或unicode)值的转换。 
    """

    @staticmethod
    def encode(value):
        """ 
        接受 basestring 子类的值( str 或 unicode )，
        否则抛出 TypeError 。
        """
        if isinstance(value, basestring):
            return value

        raise TypeError

    @staticmethod
    def decode(value):
        """ 
        将值转回str或unicode类型。 
        """
        if value is None:
            return None
        else:
            try:
                return str(value)
            except:
                return unicode(value)



# coding: utf-8

__all__ = [
    'connect', 'get_client',
    'type_case',
    'Dict', 'Set', 'SortedSet', 'String', 'Counter', 'Deque',
    '__version__',
]

import type_case

from client import connect, get_client

from key.dict import Dict
from key.set import Set
from key.sorted_set import SortedSet
from key.string import String
from key.counter import Counter
from key.deque import Deque

__version__ = "1.9.2"



# coding: utf-8

__all__ = ['Counter']

__metaclass__ = type

from ooredis import String
from ooredis.type_case import IntTypeCase
from ooredis.key.helper import wrap_exception
from ooredis.const import DEFAULT_INCREMENT, DEFAULT_DECREMENT

class Counter(String):

    """
    计算器用途的 key 对象。
   
    注意当 Key 对象为空时，get/getset 的返回值为 None。
    """

    def __init__(self, name, client=None, type_case=IntTypeCase):
        """ 
        初始化一个 Counter 类实例，
        使用 IntTypeCase 作为默认 type case 。
        """
        super(Counter, self).__init__(name=name, client=client, type_case=type_case)


    @wrap_exception
    def incr(self, increment=DEFAULT_INCREMENT):
        """
        将计数器的值加上增量 increment， 
        然后返回执行 incr 操作之后计数器的当前值。

        Args:
            increment: 增量，默认为 1 。

        Time:
            O(1)

        Returns:
            current_counter_value

        Raises:
            TypeError: 当 Key 对象储存的不是数值类型时抛出。
        """
        # redis-py用incr代替redis的incrby
        return self._client.incr(self.name, increment)


    @wrap_exception
    def decr(self, decrement=DEFAULT_DECREMENT):
        """
        将计数器的值减去减量 decrement ，
        然后返回执行 decr 操作之后计数器的当前值。

        Args:
            decrement: 减量，默认为1.

        Time:
            O(1)

        Returns:
            current_counter_value

        Raises:
            TypeError: 当 key 储存的不是数值类型时抛出。
        """
        # redis-py用decr代替redis的decrby
        return self._client.decr(self.name, decrement)


    def __iadd__(self, increment):
        """
        self.incr 方法的一个 python 语法糖，
        区别是这个特殊方法不返回 key 对象的当前值。

        Args:
            increment: 增量。

        Time:
            O(1)

        Returns:
            self: python 要求这个特殊方法返回被修改后的对象。

        Raises:
            TypeError: 当 key 储存的不是数值类型时由 self.incr 抛出。
        """
        self.incr(increment)
        return self


    def __isub__(self, decrement):
        """
        self.decr 方法的一个 python 语法糖，
        区别是这个特殊方法不返回 key 对象的当前值。

        Args:
            decrement: 减量。

        Time:
            O(1)

        Returns:
            self: python 要求这个特殊方法返回被修改后的对象。

        Raises:
            TypeError: 当 key 储存的不是数值类型时由 self.decr 抛出。
        """
        self.decr(decrement)
        return self



# coding: utf-8

from numbers import Integral
from helper import is_any_instance

class FloatTypeCase:

    """ 
    处理 float 类型值的转换。 
    """

    @staticmethod
    def encode(value):
        """
        接受 float 类型值，否则抛出 TypeError 。 
        """
        if is_any_instance(value, float, int):
            return float(value)
        
        raise TypeError

    @staticmethod
    def decode(value):
        """ 
        尝试将值转换成 float 类型，
        如果转换失败，抛出 TypeError 。
        """
        return None if value is None else float(value)



# coding: utf-8

__all__ = ['connect', 'get_client']

import redis

client = None

def connect(*args, **kwargs):
    """ 连接Redis数据库，参数和redis-py的Redis类一样 """
    global client
    client = redis.Redis(*args, **kwargs)

def get_client():
    """ 返回OORedis客户端 """
    global client

    if client == None:
        connect()

    return client



# coding: utf-8

__all__ = ['SortedSet']

__metaclass__ = type

from functools import partial

from ooredis.key.base_key import BaseKey
from ooredis.key.helper import format_key, wrap_exception
from ooredis.const import (
    LEFTMOST,
    RIGHTMOST,
    DEFAULT_INCREMENT,
    DEFAULT_DECREMENT,
)

# redis command execute status code
MEMBER_NOT_IN_SET_AND_REMOVE_FALSE = 0
MEMBER_NOT_IN_SET_AND_DELETE_FALSE = 0
MEMBER_NOT_IN_SET_AND_GET_SCORE_FALSE = None

# ZRANGE result item index
VALUE = 0
SCORE= 1

class SortedSet(BaseKey):

    """ 
    有序集对象，底层是redis的zset实现。 
    """

    def __repr__(self):
        return format_key(self, self.name, list(self))
  

    @wrap_exception
    def __len__(self):
        """ 
        返回有序集的基数。

        Args:
            None

        Time:
            O(1)

        Returns:
            int: 有序集的基数，集合不存在或为空时返回0。

        Raises：
            TypeError: 当key不是有序集类型时抛出。
        """
        return self._client.zcard(self.name)


    @wrap_exception
    def __contains__(self, element):
        """ 
        检查给定元素 element 是否是有序集的成员。 

        Args:
            element

        Time:
            O(log(N))

        Returns:
            bool: 如果 element 是集合的成员，返回 True ，否则 False 。

        Raises:
            TypeError: 当 key 不是有序集类型时抛出。
        """
        # 因为在 redis 里 zset 没有 set 那样的 SISMEMBER 命令，
        # 所以这里用 ZSCORE 命令 hack 一个：
        # 如果 ZSCORE key member 不为 None ，证明 element 是有序集成员。
        # 注意，这里不能用 self.score ，因为这两个方法互相引用。
        redis_element = self.encode(element)
        element_score = self._client.zscore(self.name, redis_element)
        return element_score is not None


    @wrap_exception
    def __setitem__(self, member, new_score):
        """ 
        将元素 member 的 score 值更新为 new_score 。

        如果 member 不存在于有序集，
        那么将 member 加入到有序集， score 值等于 new_score 。

        Args:
            member
            new_score

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当key不是有序集类型时抛出。
        """
        redis_memeber = self.encode(member)
        self._client.zadd(self.name, redis_memeber, new_score)


    @wrap_exception
    def __getitem__(self, index):
        """ 
        返回有序集指定下标内的元素。

        Args:
            index: 一个下标或一个 slice 对象。

        Returns:
            dict： 使用下标时，返回一个字典，包含 value 和 score 。
            list：使用 slice 对象时，返回一个列表，
                  列表中每个项都是一个字典。

        Time:
            O(log(N)+M) ， N 为有序集的基数，而 M 为结果集的基数。

        Raises:
            KeyError: index 下标超出范围时抛出。
            TypeError: 当 key 不是有序集类型时抛出。
        """
        item_to_dict_list = \
            partial(map,
                    lambda item: dict(
                        value=self.decode(item[VALUE]), 
                        score=item[SCORE]
                    ))

        if isinstance(index, slice):
            items = self._client.zrange(self.name, LEFTMOST, RIGHTMOST, withscores=True)
            return item_to_dict_list(items[index])
        else:
            items = self._client.zrange(self.name, index, index, withscores=True)
            return item_to_dict_list(items)[0]


    @wrap_exception
    def __delitem__(self, index):
        """ 
        删除有序集指定下标内的元素。

        Args:
            index: 下标或一个 slice 对象。

        Time:
            O(log(N)+M) ， N 为有序集的基数，而 M 为被移除成员的数量。

        Returns:
            None

        Raises:
            KeyError: index下标超出范围时抛出。
            TypeError: 当key不是有序集类型时抛出。
        """
        if isinstance(index, slice):
            start = LEFTMOST if index.start is None else index.start
            stop = RIGHTMOST if index.stop is None else index.stop-1

            self._client.zremrangebyrank(self.name, start, stop)
        else:
            status = self._client.zremrangebyrank(self.name, index, index)
            if status == MEMBER_NOT_IN_SET_AND_DELETE_FALSE:
                raise IndexError


    @wrap_exception
    def remove(self, member):
        """ 
        移除有序集成员member，如果member不存在，不做动作。

        Args:
            member: 要移除的成员。

        Time:
            O(log(N))

        Returns:
            None

        Raises:
            TypeError: 当 key 不是有序集类型时抛出。
        """
        redis_member = self.encode(member)
        self._client.zrem(self.name, redis_member)


    @wrap_exception
    def rank(self, member):
        """ 
        按从小到大的顺序(正序)返回有序集成员 member 的 score 值的排名。

        Args:
            member: 被检查的成员。

        Time:
            O(log(N))

        Returns:
            None: 当 member 不是有序集的成员时返回
            int: member 的 score 排名。

        Raises:
            TypeError: 当 key 不是有序集类型时由 in 语句抛出。
        """
        redis_member = self.encode(member)
        return self._client.zrank(self.name, redis_member)


    @wrap_exception
    def reverse_rank(self, member):
        """
        按从大到小的顺序(逆序)返回有序集成员 member 的 score 值排名。

        Args:
            member

        Time:
            O(log(N))

        Returns:
            None: 当 member 不是有序集的成员时返回
            int: member 的 score 值排名

        Raises:
            TypeError: 当 key 不是有序集类型时由 in 语句抛出。
        """
        redis_member = self.encode(member)
        return self._client.zrevrank(self.name, redis_member)


    @wrap_exception
    def score(self, member):
        """ 
        返回有序集中成员 member 的 score 值。

        Args:
            member

        Time:
            O(1)

        Returns:
            None: 当 member 不是有序集的成员时返回
            float: 以浮点值表示的 score 值。

        Raises:
            TypeError: 当 key 不是有序集类型时抛出，由 in 语句抛出。
        """
        redis_member = self.encode(member)
        return self._client.zscore(self.name, redis_member)


    @wrap_exception
    def incr(self, member, increment=DEFAULT_INCREMENT):  
        """ 
        将 member 的 score 值加上 increment 。 

        当 key 不存在，或 member 不是 key 的成员时，
        self.incr(member, increment) 等同于 self[member] = increment

        Args:
            member: 集合成员
            increment: 增量，默认为1。

        Time:
            O(log(N))

        Returns:
            float: member 成员的新 score 值。

        Raises:
            TypeError: 当 key 不是有序集类型时抛出。
        """
        redis_member = self.encode(member)
        return self._client.zincrby(self.name, redis_member, increment)


    def decr(self, member, decrement=DEFAULT_DECREMENT):
        """ 
        将 member 的 score 值减去 decrement 。 

        当 key 不存在，或 member 不是 key 的成员时，
        self.decr(member, decrement) 等同于 self[member] = 0-decrement

        Args:
            member: 集合成员
            decrement: 减量，默认为 1 。

        Time:
            O(log(N))

        Returns:
            float: member成员的新score值。

        Raises:
            TypeError: 当key不是有序集类型时抛出。
        """
        return self.incr(member, 0-decrement)



# coding:utf-8

__all__ = ['Set']

__metaclass__ = type

import redis
import collections

from ooredis.key.base_key import BaseKey
from ooredis.const import REDIS_TYPE
from ooredis.key.helper import format_key, wrap_exception

REMOVE_SUCCESS = True
MOVE_FAIL_CAUSE_MEMBER_NOT_IN_SET = 0

class Set(BaseKey):

    """
    集合 key 对象，底层实现是redis的set类型。 
    """

    def __repr__(self):
        return format_key(self, self.name, set(self))


    @wrap_exception
    def __len__(self):
        """ 
        返回集合中元素的个数。
        当集合为空集时，返回 0 。

        Args:
            None

        Time:
            O(1)

        Returns:
            len 

        Raises:
            TypeError: 当 key 不是 redis 的 set 类型时抛出。
        """
        return self._client.scard(self.name)


    def __iter__(self):
        """ 
        返回一个包含集合中所有元素的迭代器。

        Time:
            O(N)

        Returns:
            iterator

        Raises:
            TypeError: 当 key 不是 redis 的 set 类型时抛出。
        """
        try:
            for redis_member in self._client.smembers(self.name):
                python_member = self.decode(redis_member)
                yield python_member
        except redis.exceptions.ResponseError:
            raise TypeError


    @wrap_exception
    def __contains__(self, element):
        """ 
        检查给定元素 element 是否集合的成员。

        Args:
            element

        Time:
            O(1)

        Returns:
            bool: element 是集合成员的话返回 True ，否则返回 False 。

        Raises:
            TypeError: 当 key 不是 redis 的 set 类型时抛出。
        """
        redis_element = self.encode(element)
        return self._client.sismember(self.name, redis_element)


    @wrap_exception
    def add(self, element):
        """ 
        将 element 加入到集合当中。

        如果 element 已经是集合的成员，不做动作。
        如果 key 不存在，一个空集合被创建并执行 add 动作。

        Args:
            element

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当 key 非空且它不是 redis 的 set 类型时抛出。
        """
        redis_element = self.encode(element)
        self._client.sadd(self.name, redis_element)


    @wrap_exception
    def remove(self, element):
        """ 
        如果 element 是集合的成员，移除它。

        如果要移除的元素不存在，抛出 KeyError 。

        Args：
            element

        Time：
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当 key 非空且它不是 redis 的 set 类型时抛出。
            KeyError： 要移除的元素 element 不存在于集合时抛出。
        """
        redis_element = self.encode(element)
        remove_state = self._client.srem(self.name, redis_element)
        if remove_state != REMOVE_SUCCESS:
            raise KeyError


    @wrap_exception
    def pop(self):
        """ 
        移除并返回集合中任意一个成员。
        如果集合为空，抛出 KeyError 。

        Args:
            None

        Time：
            O(1)

        Returns:
            member: 被移除的集合成员。

        Raises:
            TypeError: 当 key 非空且它不是 redis 的 set 类型时抛出。
            KeyError: 集合为空集时抛出。
        """
        redis_member = self._client.spop(self.name)
        python_member = self.decode(redis_member)
        if python_member is None:
            raise KeyError
        else:
            return python_member


    @wrap_exception
    def random(self):
        """ 
        随机返回集合中的某个元素。

        该操作和 pop 相似，但 pop 将随机元素从集合中移除并返回，
        而 random 则仅仅返回随机元素，而不对集合进行任何改动。

        Args:
            None

        Time:
            O(1)

        Returns:
            member: 当集合非空时，返回一个成员。
            None: 当集合为空集时，返回 None 。

        Raises:
            TypeError: 当 key 非空且它不是 redis 的 set 类型时抛出。
        """
        redis_element = self._client.srandmember(self.name)
        python_member = self.decode(redis_element)
        return python_member


    @wrap_exception
    def move(self, destination, member):
        """ 
        将集合成员 member 移动到另一个集合 destination 中去。

        具体参考 Redis 命令：SMOVE。

        Args:
            destination: 指定被移动元素的目的地集合， 必须是一个集合 key 对象。
            member: 被移动的源集合的成员。

        Time:
            O(1)

        Returns:
            None

        Raises:
            KeyError: 要被移动的元素 member 不存在于集合时抛出。
            TypeError: 当 key 非空且它不是 redis 的 set 类型时抛出。
        """
        redis_member = self.encode(member)
        state = self._client.smove(self.name, destination.name, redis_member)
        if state == MOVE_FAIL_CAUSE_MEMBER_NOT_IN_SET:
            raise KeyError


    # disjoint, 不相交

    def isdisjoint(self, other):
        """ 
        检查集合是否和另一个集合不相交。

        Args:
            other: 一个 python 集合或集合 key 对象。
        
        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        other_member = set(other)
        self_member = set(self)

        return self_member.isdisjoint(other_member)


    # subset, <= , 子集

    def __le__(self, other):
        """ 
        测试集合是否是另一个集合的子集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) <= set(other)

    issubset = __le__


    # proper subset, < , 真子集

    def __lt__(self, other):
        """ 
        测试集合是否是另一个集合的真子集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) < set(other)


    # superset, >= , 超集

    def __ge__(self, other):
        """ 
        测试集合是否是另一个集合的超集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) >= set(other)

    issuperset = __ge__


    # proper superset, > , 真超集

    def __gt__(self, other):
        """ 
        测试集合是否是另一个集合的真超集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Returns:
            bool

        Time:
            O(N)

        Raises:
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) > set(other)


    # union, | , 并集

    def __or__(self, other):
        """ 
        返回集合和另一个集合的并集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Time:
            O(N)

        Returns:
            set

        Raises：
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) | set(other)

    __ror__ = __or__
    __ror__.__doc__ = """ __or__的反向方法，用于支持多集合对象进行并集操作。"""


    @wrap_exception
    def __ior__(self, other):
        """
        计算 self 和 other 之间的并集，并将结果设置为 self 的值。
        other 可以是另一个集合 key 对象，或者 python set 的实例。

        Args:
            other

        Time:
            O(N)

        Returns:
            self: Python 指定该方法必须返回 self 。

        Raises:
            TypeError: 当 key 或 other 的类型不符合要求时抛出。
        """
        if isinstance(other, set):
            elements = set(self) | other
            redis_elements = map(self.encode, elements)
            self._client.sadd(self.name, *redis_elements)
        else:
            self._client.sunionstore(self.name, [self.name, other.name])

        return self
 

    # intersection, & , 交集

    def __and__(self, other):
        """ 
        返回集合和另一个集合的交集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Time:
            O(N)

        Returns:
            set

        Raises：
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) & set(other)

    __rand__ = __and__
    __rand__.__doc__ = """ __and__的反向方法，用于支持多集合进行交集操作。 """


    @wrap_exception
    def __iand__(self, other):
        """
        计算 self 和 other 之间的交集，并将结果设置为 self 的值。
        other 可以是另一个集合 key 对象，或者 python set 的实例。

        Args:
            other

        Time:
            O(N)

        Returns:
            self: Python 指定该方法必须返回 self 。

        Raises:
            TypeError: 当 key 或 other 的类型不符合要求时抛出。
        """
        if isinstance(other, set):
            elements = set(self) - (set(self) & other)
            redis_elements = map(self.encode, elements)
            self._client.srem(self.name, *redis_elements)
        elif other.exists and other._represent != REDIS_TYPE['set']:
            raise TypeError
        else:
            self._client.sinterstore(self.name, [self.name, other.name])

        return self


    # difference, - , 差集

    def __sub__(self, other):
        """ 
        返回集合对另一个集合的差集。

        Args:
            other: 一个 python 集合或集合 key 对象。

        Time:
            O(N)

        Returns:
            set

        Raises：
            TypeError: 当 key 或 other 不是 Set 类型时抛出。
        """
        return set(self) - set(other)


    def __rsub__(self, other):
        """ 
        __sub__的反向方法，
        用于支持多集合进行差集运算。
        """
        # WARNING: 注意这里的位置不要弄反。
        return set(other) - set(self)


    @wrap_exception
    def __isub__(self, other):
        """
        计算 self 和 other 之间的差集，并将结果设置为 self 的值。
        other 可以是另一个集合 key 对象，或者 python set 的实例。

        Args:
            other

        Time:
            O(N)

        Returns:
            self: Python指定该方法必须返回self。

        Raises:
            TypeError: 当 key 或 other 的类型不符合要求时抛出。
        """
        if isinstance(other, set):
            remove_elements = set(self) & other
            redis_elements = map(self.encode, remove_elements)
            self._client.srem(self.name, *redis_elements)
        else:
            self._client.sdiffstore(self.name, [self.name, other.name])

        return self


    # symmetric difference, ^ ， 对等差

    def __xor__(self, other):
        """ 
        返回集合对另一个集合的对等差集。

        Args:
            other: 一个python集合或集合key对象。

        Returns:
            set

        Time:
            O(N)

        Raises:
            TypeError: 当key或other不是Set类型时抛出。
        """
        return set(self) ^ set(other)

    __rxor__ = __xor__
    __rxor__.__doc__ = \
    """ __xor__的反向方法，用于支持多集合的对等差集运算。 """



# coding: utf-8

def is_any_instance(value, *types):
    """ 接受一个值 value ，以及任意数量的类型 types ，
    如果 value 是 types 中任何一个类型的实例时，返回 True 。
    """
    return any(map(lambda t: isinstance(value, t), types))



# coding: utf-8

__all__ = ['Deque']

__metaclass__ = type

import collections
import redis.exceptions as redispy_exception

from ooredis.key.base_key import BaseKey
from ooredis.key.helper import format_key, wrap_exception

class Deque(BaseKey):

    """ 
    一个双端队列 key 对象，底层实现是 redis 的 list 类型。 
    """

    def append(self, python_item):
        """
        将元素 item 追加到队列的最右边。

        Args:
            item

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        redis_item = self.encode(python_item)
        self._client.rpush(self.name, redis_item)


    @wrap_exception
    def appendleft(self, python_item):
        """
        将元素 item 追加到队列的最左边。

        Args:
            item

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        redis_item = self.encode(python_item)
        self._client.lpush(self.name, redis_item)


    def extend(self, python_iterable):
        """
        将 iterable 内的所有元素追加到队列的最右边。

        Args:
            iterable

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        redis_iterable = map(self.encode, python_iterable)
        self._client.rpush(self.name, *redis_iterable)


    @wrap_exception
    def extendleft(self, python_iterable):
        """
        将 iterable 内的所有元素追加到队列的最左边。

        注意被追加元素是以逆序排列的。

        比如对一个空队列 d 执行 d.extendleft(range(3)) ，
        那么队列 d 将变成 [3, 2, 1] 。

        Args:
            iterable

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        redis_iterable = map(self.encode, python_iterable)
        self._client.lpush(self.name, *redis_iterable)


    def clear(self):
        """
        删除队列中的所有元素。

        Time:
            O(1)
        
        Returns:
            None

        Raises:
            None
        """
        self.delete()
    

    def count(self, python_item):
        """
        计算队列中和 item 相等的元素的个数。

        Args:
            item

        Time:
            O(N)

        Returns:
            count

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        all_python_item = list(self)
        return all_python_item.count(python_item)


    @wrap_exception
    def pop(self):
        """
        移除并返回队列最右边的元素。

        如果队列为空， 抛出 IndexError 。

        Args:
            None

        Time:
            O(1)

        Returns:
            pop_item

        Raises:
            IndexError: 尝试对空队列执行 pop 操作时抛出。
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        redis_item = self._client.rpop(self.name)
        python_item = self.decode(redis_item)
        if python_item is None:
            raise IndexError
        else:
            return python_item


    @wrap_exception
    def block_pop(self, timeout=0):
        """
        移除并返回队列最右边的元素，
        如果队列中没有元素，那么阻塞 timeout 秒，直到获取元素或超时为止。

        Args:
            timeout: 等待元素时的最大阻塞秒数

        Time:
            O(1)

        Returns:
            None: 超时时返回
            pop_item: 被弹出的元素

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        # 每个非空 brpop 的结果都是一个列表 ['list of the pop item', 'pop item']
        redis_queue_name_and_item_list = self._client.brpop(self.name, timeout)
        if redis_queue_name_and_item_list is not None:
            redis_item = redis_queue_name_and_item_list[1]
            python_item = self.decode(redis_item)
            return python_item


    @wrap_exception
    def popleft(self):
        """
        移除并返回队列最左边的元素。

        如果队列为空，抛出 IndexError 。

        Args:
            None

        Time:
            O(1)

        Returns:
            pop_item

        Raises:
            IndexError: 尝试对空队列执行 pop 操作时抛出。
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        redis_item = self._client.lpop(self.name)
        python_item = self.decode(redis_item)
        if python_item is None:
            raise IndexError
        else:
            return python_item


    @wrap_exception
    def block_popleft(self, timeout=0):
        """
        移除并返回队列最左边的元素，
        如果队列中没有元素，那么阻塞 timeout 秒，直到获取元素或超时为止。

        Args:
            timeout: 等待元素时的最大阻塞秒数

        Time:
            O(1)

        Returns:
            None: 超时时返回
            pop_item: 被弹出的元素

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        # 每个非空 blpop 的结果都是一个列表 ['list of the pop item', 'pop item']
        redis_queue_name_and_item_list = self._client.blpop(self.name, timeout)
        if redis_queue_name_and_item_list is not None:
            redis_item = redis_queue_name_and_item_list[1]
            python_item = self.decode(redis_item) 
            return python_item


    def __repr__(self):
        return format_key(self, self.name, list(self))


    @wrap_exception
    def __len__(self):
        """
        返回队列中的元素个数。
        空列表返回 0 。

        Time:
            O(1)

        Returns:
            len

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        return self._client.llen(self.name)


    def __iter__(self):
        """
        返回一个包含整个队列所有元素的迭代器。

        Time:
            O(N)

        Returns:
            iterator

        Raises:
            TypeError: 尝试对非 list 类型的 key 进行操作时抛出。
        """
        try:
            all_redis_item = self._client.lrange(self.name, 0, -1)
            for redis_item in all_redis_item:
                python_item = self.decode(redis_item)
                yield python_item
        except redispy_exception.ResponseError:
            raise TypeError


    def __delitem__(self, key):
        raise Exception 


    def __getitem__(self, index):
        all_item = list(self)
        return all_item[index]


    def __setitem__(self, index, item):
        raise Exception



# coding: utf-8

from numbers import Integral

class IntTypeCase:

    """ 
    处理 int(long) 类型值的转换。
    """

    @staticmethod
    def encode(value):
        """ 
        接受 int 类型值，否则抛出 TypeError 。 
        """
        if isinstance(value, Integral):
            return value

        raise TypeError

    @staticmethod
    def decode(value):
        """ 
        尝试将值转回 int 类型，
        如果转换失败，抛出 TypeError。
        """
        if value is None:
            return None
        else:
            try:
                return int(value)
            except:
                try:
                    return long(value)
                except:
                    raise TypeError



from int_type_case import IntTypeCase
from json_type_case import JsonTypeCase
from float_type_case import FloatTypeCase
from string_type_case import StringTypeCase
from generic_type_case import GenericTypeCase
from serialize_type_case import SerializeTypeCase



# coding:utf-8

__all__ = ['Dict']

__metaclass__ = type

import collections
import redis.exceptions as redispy_exception

from ooredis.key.base_key import BaseKey
from ooredis.key.helper import format_key, wrap_exception

DELETE_FAIL_CAUSE_KEY_NOT_EXISTS = False

class Dict(BaseKey, collections.MutableMapping):

    """
    一个字典对象，底层是 Redis 的 Hash 结构。
    """

    def __repr__(self):
        return format_key(self, self.name, dict(self))


    @wrap_exception
    def __setitem__(self, key, python_value):
        """ 
        将字典中键 key 的值设为 python_value 。
        如果键 key 已经关联了另一个值 ，那么将它覆盖。

        Args:
            key
            python_value

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: Key 对象不是 Dict 类型时抛出。
        """
        redis_value = self.encode(python_value)
        self._client.hset(self.name, key, redis_value)


    @wrap_exception
    def __getitem__(self, key):
        """ 
        返回字典中键 key 的值。
        如果键 key 在字典中不存在，那么抛出 KeyError 。

        Args:
            key

        Time:
            O(1)

        Returns:
            python_value

        Raises:
            KeyError: key 不存在时抛出。
            TypeError: Key 对象不是 Dict 类型时抛出。
        """
        # 没有自己实现 __contains__ 的话
        # 不要使用 key in self ，否则将引起死循环
        if not self._client.hexists(self.name, key):
            raise KeyError

        redis_value = self._client.hget(self.name, key)
        python_value = self.decode(redis_value)

        return python_value

    
    @wrap_exception
    def __delitem__(self, key):
        """ 
        删除字典键 key 的值。
        如果键 key 的值不存在，那么抛出 KeyError 。

        Args:
            key

        Time:
            O(1)

        Returns:
            None

        Raises:
            KeyError: key 不存在时抛出。
            TypeError: Key 对象不是 Dict 类型时抛出。
        """
        status = self._client.hdel(self.name, key)
        if status == DELETE_FAIL_CAUSE_KEY_NOT_EXISTS:
            raise KeyError


    # @wrap_exception
    # 似乎因为 yield 的缘故，装饰器没办法在这里使用
    def __iter__(self):
        """ 
        返回一个包含字典里所有键的迭代器。

        Args:
            None

        Time:
            O(N)

        Returns:
            iterator
        
        Raises:
            TypeError: Key 对象不是 Dict 类型时抛出。
        """
        try:
            for key in self._client.hkeys(self.name):
                yield key
        except redispy_exception.ResponseError:
            raise TypeError

   
    @wrap_exception
    def __len__(self):
        """
        返回字典中键-值对的个数。
        空字典返回 0 。

        Args:
            None

        Time:
            O(1)

        Returns:
            len

        Raises:
            TypeError: Key 对象不是 Dict 类型时抛出。
        """
        return self._client.hlen(self.name)



# coding: utf-8

from helper import is_any_instance
from numbers import Integral

class GenericTypeCase:

    """ 
    通用类型转换类，可以处理 str/unicode/int/float 类型值。
    """

    @staticmethod
    def encode(value):
        """ 
        接受 int/long/str/unicode/float 类型值，
        否则抛出 TypeError 。
        """
        if is_any_instance(value, basestring, Integral, float):
            return value

        raise TypeError

    @staticmethod
    def decode(value):
        """ 
        将传入值转换成 int/long/str/unicode/float 等格式。
        因为 redis 只返回字符串值，这个函数也只接受字符串值，
        否则抛出 TypeError 。
        """
        if value is None:
            return None
        elif not isinstance(value, basestring):
            # redis 返回的只能是字符串值
            raise TypeError
        else:
            # 从最限定的类型开始转换
            try: return int(value)
            except:
                try: return float(value)
                except:
                    try: return str(value)
                    except:
                        return unicode(value)



# coding: utf-8

import pickle

class SerializeTypeCase:

    """ 
    处理 python 对象序列化(使用 pickle 模块)。
    """

    @staticmethod
    def encode(value):
        """ 
        将值序列化(包括自定义类和 Python 内置类)。
        如果传入的值无法被序列化，抛出 TypeError 。
        """
        try:
            return pickle.dumps(value)
        except pickle.PicklingError:
            raise TypeError

    @staticmethod
    def decode(value):
        """ 
        将序列化的字符串转换回原来的对象，
        如果传入值无法进行反序列化，抛出 TypeError 。
        """
        if value is None:
            return None

        try:
            return pickle.loads(value)
        except pickle.UnpicklingError:
            raise TypeError



# coding:utf-8

__all__ = [
    'format_key',
    'wrap_exception',
]

import redis

def format_key(key, name, value):
    """ 
    提供对 Key 对象的格式化支持。
    """
    type = key.__class__.__name__.title()
    return "{0} Key '{1}': {2}".format(type, name, value)

def wrap_exception(func):
    """
    redis-py 在对数据结构执行不属于它的操作时抛出 ResponseError
    这里将 ResponseError 转换为 TypeError
    """
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except redis.exceptions.ResponseError:
            raise TypeError
    return wrapper



# coding: utf-8

__all__ = ['String']

__metaclass__ = type

from functools import wraps
from ooredis.key.base_key import BaseKey
from ooredis.const import REDIS_TYPE
from ooredis.key.helper import format_key, wrap_exception


def raise_when_set_wrong_type(func):
    """
    set/setex/setnx 命令可以无视类型进行设置的命令，为了保证类型的限制，
    ooredis 对一个非 string 类型进行 set/setex/setnx 将引发 TypeError 异常。
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        if self.exists and \
           self._represent != REDIS_TYPE['string']:
            raise TypeError
        return func(*args, **kwargs)
    return wrapper


class String(BaseKey):

    """
    为储存单个值的 Key 对象提供 set，get 和 getset操作。
    """

    def __repr__(self):
        return format_key(self, self.name, self.get())


    @raise_when_set_wrong_type
    def set(self, python_value):
        """
        为 Key 对象指定值。

        Args:
            python_value: 为 Key 对象指定的值。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当 key 非空但 Key 对象不是指定类型时抛出。
        """
        redis_value = self.encode(python_value)

        self._client.set(self.name, redis_value)


    @raise_when_set_wrong_type
    def setnx(self, python_value):
        """
        将 Key 对象的值设为 python_value ，当且仅当 Key 不存在。

        Args:
            python_value

        Time:
            O(1)

        Returns:
            bool: 如果设置成功，那么返回 True ，否则返回 False 。

        Raises:
            TypeError: 当 Key 对象非空且不是 Redis 的字符串类型时抛出。
        """
        redis_value = self.encode(python_value)

        return self._client.setnx(self.name, redis_value)


    @raise_when_set_wrong_type
    def setex(self, python_value, ttl_in_second):
        """
        将 Key 对象的值设为 python_value ，
        并将 Key 对象的生存时间设为 ttl_in_second 。

        Args:
            python_value
            ttl_in_second: 生存时间，以秒为单位。

        Time:
            O(1)

        Returns:
            None

        Raises:
            TypeError: 当 Key 对象非空且不是 Redis 的字符串类型时抛出。
        """
        redis_value = self.encode(python_value)

        return self._client.setex(self.name, redis_value, ttl_in_second)


    @wrap_exception
    def get(self):
        """
        返回 Key 对象的值。

        Time:
            O(1)

        Returns:
            python_value： Key 对象的值。
            None: key 不存在时返回。

        Raises:
            TypeError: 当 key 非空但 Key 对象不是指定类型时抛出。
        """
        redis_value = self._client.get(self.name)

        python_value = self.decode(redis_value)

        return python_value

   
    @wrap_exception
    def getset(self, python_value):
        """ 
        修改 Key 对象的值，并返回 Key 对象之前储存的值。

        Args:
            python_value: Key 对象的新值。

        Time:
            O(1)

        Returns:
            None: 当 key 不存时(没有前值)返回。
            python_value:  Key 对象之前的值。

        Raises:
            TypeError: 当 key 非空但 Key 对象不是指定类型时抛出。
        """
        new_redis_value = self.encode(python_value)

        old_redis_value = self._client.getset(self.name, new_redis_value)

        python_value = self.decode(old_redis_value)

        return python_value
